MODULE FreeOberon;

IMPORT G := Graph, T := Terminal, Text := EditorText, Widgets, Term, Config,
       Strings, Out;

CONST
  needFullscreen = TRUE;
  version* = '0.9';

  (* Direction of Selection *)
  dirLeft  = 0;
  dirRight = 1;
  dirUp    = 2;
  dirDown  = 3;
  
  (* States *)
  stateEditor   = 0;
  stateTerminal = 1;

  (* Character Kinds *)
  charOther       = 0;
  charAlpha       = 1;
  charDigit       = 2;
  charMinusPlus   = 3;
  charQuote       = 4;
  charOpenBracket = 5;

  (* Token Kinds *)
  tokenOther   = 0;
  tokenKeyword = 1;
  tokenNumber  = 2;
  tokenString  = 3;
  tokenComment = 4;

VAR
  keys: G.KeyArray;
  text: Text.Text;
  fileIsNew: BOOLEAN; (* True if file has not yet been saved *)
  fileName: ARRAY 1024 OF CHAR;
  editor: Widgets.EditingWindow;
  inputBox: Widgets.Edit;
  focusedWidget: Widgets.Widget;
  saveAs: BOOLEAN;
  state: INTEGER;
  clipboard: ARRAY 16300 OF CHAR;
  
  progBuf: ARRAY 16300 OF CHAR; (* For interacting with a launched program *)
  inputBuf: ARRAY 16300 OF CHAR; (* Saves entered characters before Enter is pressed *)
  inputBufLen: INTEGER;
  programFinished: BOOLEAN;
  isFullscreen: BOOLEAN;
  tempWindowed: BOOLEAN; (* True if editor is in windowed mode while program is running *)

PROCEDURE IntToStr*(n: INTEGER; VAR s: ARRAY OF CHAR); (* !TODO move out ; *)
(* LEN(s) > 1 *)
VAR i, j: INTEGER; tmp: CHAR; neg: BOOLEAN;
BEGIN
  IF n = 0 THEN
    s[0] := '0'; i := 1
  ELSE i := 0; neg := n < 0; 
    IF neg THEN n := -n END;
    WHILE (n > 0) & (i < LEN(s) - 1) DO
      s[i] := CHR(ORD('0') + n MOD 10);
      n := n DIV 10; INC(i)
    END;
    IF neg & (i < LEN(s) - 1) THEN s[i] := '-'; INC(i) END;
  END;
  s[i] := 0X; j := 0; DEC(i);
  WHILE j < i DO
    tmp := s[j]; s[j] := s[i]; s[i] := tmp;
    INC(j); DEC(i)
  END
END IntToStr;

PROCEDURE StringsEqual(a, b: ARRAY OF CHAR): BOOLEAN; (* !TODO move out *)
(** Returns true if strings are equal.*)
VAR i: INTEGER;
BEGIN i := 0;
  WHILE (i < LEN(a)) & (i < LEN(b)) &
        (a[i] # 0X) & (a[i] = b[i])
  DO INC(i) END;
  RETURN (i < LEN(a)) & (i < LEN(b)) & (a[i] = 0X) & (b[i] = 0X)
END StringsEqual;

PROCEDURE StringsFindNext* (pattern, stringToSearch: ARRAY OF CHAR; startPos: INTEGER;
                     VAR patternFound: BOOLEAN; VAR posOfPattern: INTEGER); (* !TODO move out *)
VAR patternPos: INTEGER;
BEGIN
  IF (startPos < Strings.Length (stringToSearch)) THEN
    patternPos := 0;
    LOOP
      IF (pattern[patternPos] = 0X) THEN     
        (* reached end of pattern *)
        patternFound := TRUE;
        posOfPattern := startPos-patternPos;
        EXIT
      ELSIF (stringToSearch[startPos] = 0X) THEN 
        (* end of string (but not of pattern) *)
        patternFound := FALSE;
        EXIT
      ELSIF (stringToSearch[startPos] = pattern[patternPos]) THEN  
        (* characters identic, compare next one *)
        INC (startPos); 
        INC (patternPos)
      ELSE                               
        (* difference found: reset indices and restart *)
        startPos := startPos-patternPos+1; 
        patternPos := 0
      END
    END
  ELSE
    patternFound := FALSE
  END
END StringsFindNext;
  
PROCEDURE StringsFindPrev* (pattern, stringToSearch: ARRAY OF CHAR; startPos: INTEGER;
                     VAR patternFound: BOOLEAN; VAR posOfPattern: INTEGER); (* !TODO move out *)
VAR patternPos, stringLength, patternLength: INTEGER;
BEGIN
  (* correct `startPos' if it is larger than the possible searching range *)
  stringLength := Strings.Length (stringToSearch);
  patternLength := Strings.Length (pattern);
  IF (startPos > stringLength-patternLength) THEN
    startPos := stringLength-patternLength
  END;
  
  IF (startPos >= 0) THEN
    patternPos := 0;
    LOOP
      IF (pattern[patternPos] = 0X) THEN     
        (* reached end of pattern *)
        patternFound := TRUE;
        posOfPattern := startPos-patternPos;
        EXIT
      ELSIF (stringToSearch[startPos] # pattern[patternPos]) THEN
        (* characters differ: reset indices and restart *)
        IF (startPos > patternPos) THEN
          startPos := startPos-patternPos-1;
          patternPos := 0
        ELSE
          (* reached beginning of `stringToSearch' without finding a match *)
          patternFound := FALSE;
          EXIT
        END
      ELSE  (* characters identic, compare next one *)
        INC (startPos); 
        INC (patternPos)
      END
    END
  ELSE
    patternFound := FALSE
  END
END StringsFindPrev;

PROCEDURE PrintIDE;
VAR i: INTEGER; s, sN: ARRAY 80 OF CHAR;
BEGIN
  T.ClearScreen;
  FOR i := 0 TO T.charsX - 1 DO
    T.PutChar(i, 0, ' ', 0, 7); (* Top Panel *)
    T.PutChar(i, T.charsY - 1, ' ', 0, 7) (* Bottom Panel *)
  END;
  T.PutString(2, 0, 'Free Oberon', 0, 7, 0);
  T.PutChar(2, 0, 'F', 4, 7); T.PutChar(7, 0, 'O', 4, 7);

  FOR i := 1 TO T.charsX - 2 DO
    T.PutChar(i, 1, 0CDX, 15, 1); (* = *)
    T.PutChar(i, T.charsY - 2, 0CDX, 15, 1)
  END;
  FOR i := 2 TO T.charsY - 3 DO
    T.PutChar(0, i, 0BAX, 15, 1); (* || *)
    T.PutChar(T.charsX - 1, i, 0B1X, 1, 3)
  END;
  T.PutChar(T.charsX - 1, 2, 01EX, 1, 3); (* Vertical Scrollbar *)
  T.PutChar(T.charsX - 1, T.charsY - 3, 01FX, 1, 3);
  T.PutChar(0, 1, 0C9X, 15, 1); (* Corners *)
  T.PutChar(T.charsX - 1, 1, 0BBX, 15, 1);
  T.PutChar(0, T.charsY - 2, 0C8X, 15, 1);
  T.PutChar(T.charsX - 2, T.charsY - 2, 0C4X, 10, 1);
  T.PutChar(T.charsX - 1, T.charsY - 2, 0D9X, 10, 1);

  s := ' ';
  Strings.Append(fileName, s);
  Strings.Append(' ', s);
  T.PutString((T.charsX - Strings.Length(fileName)) DIV 2, 1, s, 15, 1, 0);
  T.PutString(2, 1, '[ ]', 15, 1, 0);
  T.PutChar(3, 1, 0FEX, 10, 1);
  T.PutString(T.charsX - 5, 1, '[ ]', 15, 1, 0);
  T.PutChar(T.charsX - 4, 1, 012X, 10, 1);

  s := ' '; (* Line and Column Number *)
  IntToStr(text.y + 1, sN);
  Strings.Append(sN, s);
  Strings.Append(':', s);
  IntToStr(text.x + 1, sN);
  Strings.Append(sN, s);
  Strings.Append(' ', s);
  T.PutString(8, T.charsY - 2, s, 15, 1, 0);
  T.PutString(1, T.charsY - 1, 'F1', 4, 7, 0);
  T.PutString(4, T.charsY - 1, 'Help', 0, 7, 0);

  IF Config.isWindows THEN s := 'WINDOWS' ELSE s := 'LINUX' END;
  Strings.Append(' v', s); Strings.Append(version, s);
  T.PutString(T.charsX - Strings.Length(s) - 1, T.charsY - 1, s, 8, 7, 0)
END PrintIDE;

PROCEDURE GetCharKind(ch: CHAR): INTEGER;
VAR kind: INTEGER;
BEGIN
  CASE ch OF
    'a'..'z', 'A'..'Z': kind := charAlpha
  | '0'..'9': kind := charDigit
  | '-', '+': kind := charMinusPlus
  | '"', "'": kind := charQuote
  | '(': kind := charOpenBracket
  ELSE kind := charOther END;
  RETURN kind
END GetCharKind;

PROCEDURE IsHexDigit(ch: CHAR): BOOLEAN;
VAR result: BOOLEAN;
BEGIN
  CASE ch OF
    'A'..'F': result := TRUE
  ELSE result := FALSE END;
  RETURN result
END IsHexDigit;

PROCEDURE IsKeywordInString(s: ARRAY OF CHAR; x, len: INTEGER): BOOLEAN;
VAR part: ARRAY 32 OF CHAR; result: BOOLEAN;
  PROCEDURE KW(keyword: ARRAY OF CHAR): BOOLEAN;
  BEGIN
    RETURN StringsEqual(keyword, part)
  END KW;
BEGIN
  CASE s[x] OF
    'A'..'Z':
    Strings.Extract(s, x, len, part);
    result := KW('ABS') OR KW('ASH') OR KW('BOOLEAN') OR KW('CAP') OR
      KW('CHAR') OR KW('CHR') OR KW('COPY') OR KW('DEC') OR KW('ENTIER') OR
      KW('EXCL') OR KW('FALSE') OR KW('HALT') OR KW('INC') OR KW('INCL') OR
      KW('INTEGER') OR KW('LEN') OR KW('LONG') OR KW('INTEGER') OR
      KW('LONGREAL') OR KW('MAX') OR KW('MIN') OR KW('NEW') OR KW('ODD') OR
      KW('ORD') OR KW('REAL') OR KW('SET') OR KW('SHORT') OR
      KW('SHORTINT') OR KW('SIZE') OR KW('TRUE') OR KW('ARRAY') OR
      KW('BEGIN') OR KW('BY') OR KW('CASE') OR KW('DIV') OR KW('DO') OR
      KW('ELSIF') OR KW('END') OR KW('EXIT') OR KW('FOR') OR KW('IF') OR
      KW('IMPORT') OR KW('IN') OR KW('IS') OR KW('LOOP') OR
      KW('MODULE') OR KW('NIL') OR KW('OR') OR KW('POINTER') OR
      KW('PROCEDURE') OR KW('RECORD') OR KW('REPEAT') OR KW('RETURN') OR
      KW('THEN') OR KW('TO') OR KW('TYPE') OR KW('VAR') OR KW('WHILE') OR
      KW('ELSE') OR KW('OF') OR KW('WITH') OR KW('LONGSET') OR
      KW('UNTIL') OR KW('CONST') OR KW('MOD')
  ELSE result := FALSE END;
  RETURN result
END IsKeywordInString;

PROCEDURE GetToken(s: ARRAY OF CHAR; x: INTEGER;
  VAR kind, len, comLevel: INTEGER);

VAR i, chKind: INTEGER;

  PROCEDURE TryNumber;
  VAR ok, hex, hexEnding, point, scale,
      scaleSign, scaleNum, finish: BOOLEAN;
  BEGIN
    IF (x = 0) OR
       (~(GetCharKind(s[x - 1]) IN {charDigit, charAlpha}) &
        (s[x - 1] # '.')) THEN
      ok := TRUE; finish := FALSE; point := FALSE;
      hex := FALSE; hexEnding := FALSE;
      scale := FALSE; scaleSign := FALSE; scaleNum := FALSE;
      REPEAT
        CASE s[i] OF
          '0'..'9': IF scale THEN scaleNum := TRUE END
        | 'D', 'E':
          IF point THEN
            IF scale THEN ok := FALSE
            ELSE scale := TRUE END
          ELSE hex := TRUE END
        | 'A', 'B', 'C', 'F':
          IF point THEN ok := FALSE ELSE hex := TRUE END
        | 'H', 'X':
          IF point OR hexEnding THEN ok := FALSE
          ELSE hexEnding := TRUE END
        | '.': IF point OR hex THEN ok := FALSE ELSE point := TRUE END
        | '-', '+':
          IF hexEnding OR ~scale OR scaleNum THEN finish := TRUE
          ELSE scaleSign := TRUE END
        ELSE
          IF point & scale & ~scaleNum THEN ok := FALSE
          ELSE finish := TRUE END
        END;
        INC(i)
      UNTIL ~ok OR finish;
      IF ok & (~hex OR hexEnding) THEN
        IF GetCharKind(s[i - 1]) # charAlpha THEN
          len := i - x - 1;
          kind := tokenNumber
        END
      END
    END
  END TryNumber;

  PROCEDURE TryString;
  VAR quote: CHAR;
  BEGIN
    quote := s[x];
    WHILE (s[i] # 0X) & (s[i] # quote) DO INC(i) END;
    IF s[i] # 0X THEN
      len := i - x + 1;
      kind := tokenString
    END
  END TryString;

  PROCEDURE TryComment;
  BEGIN
    IF (i > 0) & (s[i] = '*') THEN INC(i); INC(comLevel) END;
    IF comLevel > 0 THEN
      REPEAT
        WHILE (s[i] # 0X) & (s[i] # '*') DO INC(i) END;
        IF s[i] = '*' THEN
          IF (i > 0) & (s[i - 1] = '(') THEN INC(comLevel)
          ELSIF s[i + 1] = ')' THEN DEC(comLevel) END;
          INC(i)
        END
      UNTIL (s[i] = 0X) OR (comLevel <= 0);
      len := i - x + 1;
      kind := tokenComment
    END
  END TryComment;

BEGIN
  kind := tokenOther; len := 0;
  IF s[x] # 0X THEN
    IF (x = 0) & (comLevel > 0) THEN i := 0; TryComment END;
    IF (s[x] # 0X) & (kind # tokenComment) THEN
      i := x + 1;
      chKind := GetCharKind(s[x]);
      CASE chKind OF
        charAlpha:
        IF (x = 0) OR (GetCharKind(s[x - 1]) # charDigit) THEN
          WHILE GetCharKind(s[i]) = charAlpha DO INC(i) END;
          len := i - x;
          IF IsKeywordInString(s, x, len) THEN kind := tokenKeyword END
        END
      | charDigit: TryNumber
      | charQuote: TryString
      | charOpenBracket: TryComment
      ELSE len := 1 END
    END
  END
END GetToken;

PROCEDURE GetTokenColor(kind: INTEGER): INTEGER;
VAR color: INTEGER;
BEGIN
  CASE kind OF
    tokenKeyword: color := 15
  | tokenNumber:  color := 3
  | tokenString:  color := 11
  | tokenComment: color := 7
  ELSE color := 14 END;
  RETURN color
END GetTokenColor;

(* Like T.PutString, but highlighted *)
PROCEDURE PutStringH*(x, y: INTEGER; s: ARRAY OF CHAR;
  bg, limit: INTEGER; VAR comLevel: INTEGER);
VAR i, fg, kind, len: INTEGER;
BEGIN
  IF limit = 0 THEN limit := T.charsX END; 
  GetToken(s, 0, kind, len, comLevel);
  fg := GetTokenColor(kind); i := 0;
  WHILE (s[i] # 0X) & (x < limit) DO
    T.PutChar(x, y, s[i], fg, bg);
    INC(i); INC(x); DEC(len);
    IF len <= 0 THEN
      GetToken(s, i, kind, len, comLevel);
      fg := GetTokenColor(kind)
    END
  END
END PutStringH;

(* Put selected string (highlighted) *)
PROCEDURE PutSelStringH*(x, y: INTEGER; s: ARRAY OF CHAR;
  bg, limit: INTEGER; VAR comLevel: INTEGER;
  fgSel, bgSel: INTEGER; x0, x1: INTEGER);
VAR i, fg, kind, len: INTEGER; sel: BOOLEAN;
BEGIN
  sel := FALSE;
  GetToken(s, 0, kind, len, comLevel);
  fg := GetTokenColor(kind); i := 0;
  WHILE (s[i] # 0X) & (x < limit) DO
    IF i = x1 + 1 THEN sel := FALSE
    ELSIF i = x0 THEN sel := TRUE END;
    IF sel THEN T.PutChar(x, y, s[i], fgSel, bgSel)
    ELSE T.PutChar(x, y, s[i], fg, bg) END;
    INC(i); INC(x); DEC(len);
    IF len <= 0 THEN
      GetToken(s, i, kind, len, comLevel);
      fg := GetTokenColor(kind)
    END
  END
END PutSelStringH;

PROCEDURE PrintText;
VAR L: Text.Line; scrY: INTEGER; textY, x0, x1, comLevel: INTEGER;
BEGIN
  PrintIDE;
  L := text.scrFirst;  scrY := 2;  textY := text.scrY;
  comLevel := L.comLevel;
  WHILE (L # NIL) & (scrY < T.charsY - 2) DO
    IF text.selected & (text.selT <= textY) & (textY <= text.selB) THEN
      IF textY = text.selT THEN x0 := text.selL
      ELSE x0 := 0 END;
      IF textY = text.selB THEN x1 := text.selR - 1
      ELSE x1 := L.len END;
      PutSelStringH(1, scrY, L.s, 1, T.charsX - 1, comLevel, 1, 7, x0, x1)
    ELSE
      PutStringH(1, scrY, L.s, 1, T.charsX - 1, comLevel)
    END;
    IF (L = text.cur) & (text.x + 1 < T.charsX)
    THEN T.GoToXY(SHORT(text.x) + 1, scrY) END; (* !FIXME x < w *)
    L := L.next; INC(scrY); INC(textY);
    IF L # NIL THEN L.comLevel := comLevel END
  END
END PrintText;

PROCEDURE StartSelection;
BEGIN
  text.selected := TRUE;
  text.selL := text.x;  text.selR := text.x;
  text.selT := text.y;  text.selB := text.y
END StartSelection;

PROCEDURE CheckSelBorders(VAR onLeft, onRight: BOOLEAN);
BEGIN
  onLeft := (text.x = text.selL) & (text.y = text.selT);
  onRight := (text.x = text.selR) & (text.y = text.selB)
END CheckSelBorders;

PROCEDURE Swap(VAR a, b: INTEGER);
VAR tmp: INTEGER;
BEGIN
  tmp := a; a := b; b := tmp
END Swap;

PROCEDURE Order(VAR a, b: INTEGER);
VAR tmp: INTEGER;
BEGIN
  IF a > b THEN tmp := a; a := b; b := tmp END
END Order;

(* Starts, removes, upgrades selection if Shift held. *)
(* dir is one of Direction of Selection constants. *)
PROCEDURE HandleSelection(dir: INTEGER; viaHoriz: BOOLEAN);
VAR mod: SET; onLeft, onRight: BOOLEAN;
BEGIN
  IF keys[G.kLShift] OR keys[G.kRShift] THEN
    CheckSelBorders(onLeft, onRight);
    IF ~text.selected OR (~onLeft & ~onRight) THEN
      StartSelection; (* Reset/new selection *)
      CheckSelBorders(onLeft, onRight)
    END;
    CASE dir OF
      dirRight:
      IF onRight THEN INC(text.selR)
      ELSIF onLeft THEN INC(text.selL) END
    | dirLeft:
      IF onLeft THEN DEC(text.selL)
      ELSIF onRight THEN DEC(text.selR) END
    | dirUp:
      IF onLeft THEN
        DEC(text.selT);
        IF viaHoriz THEN text.selL := text.cur.prev.len
        ELSIF text.selL > text.cur.prev.len
        THEN text.selL := text.cur.prev.len END
      ELSIF viaHoriz THEN
        DEC(text.selB); text.selR := text.cur.prev.len
      ELSIF (text.selB > text.selT + 1) OR
         ((text.selB = text.selT + 1) &
          (text.selR >= text.selL)) THEN
        DEC(text.selB);
        IF text.selR > text.cur.prev.len
        THEN text.selR := text.cur.prev.len END
      ELSIF text.selT = text.selB THEN
        DEC(text.selT); Order(text.selR, text.selL);
        IF text.selL > text.cur.prev.len
        THEN text.selL := text.cur.prev.len END
      ELSE
        DEC(text.selB); Order(text.selL, text.selR);
        IF text.selR > text.cur.prev.len
        THEN text.selR := text.cur.prev.len END
      END
    | dirDown:
      IF onRight THEN
        INC(text.selB);
        IF viaHoriz THEN text.selR := 0
        ELSIF text.selR > text.cur.next.len
        THEN text.selR := text.cur.next.len END
      ELSIF viaHoriz THEN
        INC(text.selT); text.selL := 0
      ELSIF (text.selB > text.selT + 1) OR
         ((text.selB = text.selT + 1) &
          (text.selR >= text.selL)) THEN
        INC(text.selT);
        IF text.selL > text.cur.next.len
        THEN text.selL := text.cur.next.len END
      ELSIF text.selT = text.selB THEN
        INC(text.selB); Swap(text.selL, text.selR);
        IF text.selR > text.cur.next.len
        THEN text.selR := text.cur.next.len END
      ELSE
        INC(text.selT); Swap(text.selL, text.selR);
        IF text.selL > text.cur.next.len
        THEN text.selL := text.cur.next.len END
      END
    ELSE
    END;
    IF ((text.selT >= text.selB) & (text.selL >= text.selR)) OR
       (text.selT > text.selB) THEN
      text.selected := FALSE
    END
  END
END HandleSelection;

PROCEDURE MoveScreenByLine(down: BOOLEAN);
BEGIN
  IF down THEN
    ASSERT(text.scrFirst.next # NIL, 98);
    text.scrFirst := text.scrFirst.next;  INC(text.scrY)
  ELSE (* Up *)
    ASSERT(text.scrFirst.prev # NIL, 98);  ASSERT(text.scrY > 0, 99);
    text.scrFirst := text.scrFirst.prev;  DEC(text.scrY)
  END;
  T.ResetCursorBlink
END MoveScreenByLine;

PROCEDURE MoveScreen(delta: INTEGER);
BEGIN
  IF delta > 0 THEN
    WHILE (delta > 0) & (text.scrFirst.next # NIL) DO
      MoveScreenByLine(TRUE); DEC(delta)
    END
  ELSE (* Up *)
    WHILE (delta < 0) & (text.scrFirst.prev # NIL) DO
      MoveScreenByLine(FALSE); INC(delta)
    END
  END
END MoveScreen;

(* Moves input cursor up and down *)
PROCEDURE MoveByLine(down, viaHoriz: BOOLEAN);
VAR moved: BOOLEAN;
BEGIN
  moved := FALSE;
  IF down THEN
    IF text.cur.next # NIL THEN
      HandleSelection(dirDown, viaHoriz);
      moved := TRUE;
      text.cur := text.cur.next;  INC(text.y);
      IF text.y >= text.scrY + T.scrH THEN MoveScreenByLine(TRUE) END
    END
  ELSE (* Up *)
    IF text.cur.prev # NIL THEN
      HandleSelection(dirUp, viaHoriz);
      moved := TRUE;
      IF text.scrFirst = text.cur THEN MoveScreenByLine(FALSE) END;
      text.cur := text.cur.prev;  DEC(text.y)
    END
  END;
  IF moved THEN
    IF text.x > text.cur.len THEN text.x := text.cur.len END;
    PrintText
  END;
  T.ResetCursorBlink
END MoveByLine;

(* Moves input cursor up and down by page *)
PROCEDURE MoveByPage(down: BOOLEAN);
VAR i: INTEGER; moved: BOOLEAN;
BEGIN
  i := 1;
  IF down THEN
    moved := text.cur.next # NIL;
    WHILE (i < T.scrH) & (text.cur.next # NIL) DO
      text.scrFirst := text.scrFirst.next;  INC(text.scrY);
      MoveByLine(TRUE, FALSE);
      INC(i)
    END
  ELSE (* Up *)
    moved := text.cur.prev # NIL;
    WHILE (i < T.scrH) & (text.cur.prev # NIL) DO
      IF text.scrY > 0 THEN text.scrFirst := text.scrFirst.prev; DEC(text.scrY) END;
      MoveByLine(FALSE, FALSE);
      INC(i)
    END
  END;
  T.ResetCursorBlink;
  IF moved THEN PrintText END
END MoveByPage;

(* Moves input cursor left and right by one char *)
PROCEDURE MoveInLine(right: BOOLEAN);
BEGIN
  IF right THEN
    IF text.x < text.cur.len THEN
      HandleSelection(dirRight, FALSE); INC(text.x)
    ELSIF text.cur.next # NIL THEN
      MoveByLine(TRUE, TRUE);
      text.x := 0;
    END
  ELSE (* Left *)
    IF text.x > 0 THEN
      HandleSelection(dirLeft, FALSE); DEC(text.x)
    ELSIF text.cur.prev # NIL THEN
      MoveByLine(FALSE, TRUE);
      text.x := text.cur.len;
    END
  END;
  T.ResetCursorBlink;
  PrintText (*!FIXME*)
END MoveInLine;

(* Moves input cursor to start and to end *)
PROCEDURE MoveToLineEdge(end: BOOLEAN);
VAR shift, onLeft, onRight: BOOLEAN;
BEGIN
  shift := keys[G.kLShift] OR keys[G.kRShift];
  IF shift THEN
    IF ~text.selected THEN StartSelection
    ELSE
      CheckSelBorders(onLeft, onRight);
      IF ~onLeft & ~onRight THEN StartSelection END
    END;
    CheckSelBorders(onLeft, onRight);
    IF end THEN
      IF onRight THEN text.selR := text.cur.len
      ELSE text.selL := text.cur.len END
    ELSE (* Home *)
      IF onRight THEN text.selR := 0
      ELSE text.selL := 0 END
    END;
    IF text.selT = text.selB THEN Order(text.selL, text.selR) END
  END;
  IF end THEN
    IF text.x < text.cur.len THEN text.x := text.cur.len END
  ELSE (* Home *)
    IF text.x > 0 THEN text.x := 0 END
  END;
  T.ResetCursorBlink;
  PrintText
END MoveToLineEdge;

PROCEDURE DeleteText;
BEGIN
  text.DeleteSelection;
  PrintText
END DeleteText;

PROCEDURE CopyText;
BEGIN
  text.CopySelection(clipboard)
END CopyText;

PROCEDURE CutText;
BEGIN
  CopyText;
  DeleteText
END CutText;

PROCEDURE PasteText;
BEGIN
  IF clipboard[0] # 0X THEN
    text.selected := FALSE;
    text.Insert(clipboard, FALSE);
    IF text.y >= text.scrY + T.scrH THEN
      MoveScreen(text.y - T.scrH + 1 - text.scrY)
    END;
    PrintText
  END
END PasteText;

PROCEDURE HandleBackspace;
BEGIN
  text.HandleBackspace;
  text.selected := FALSE;
  PrintText
END HandleBackspace;

PROCEDURE HandleDelete;
BEGIN
  text.HandleDelete;
  PrintText
END HandleDelete;

PROCEDURE HandleEnter;
BEGIN
  text.HandleEnter(TRUE);
  IF text.y >= text.scrY + T.scrH THEN MoveScreenByLine(TRUE) END;
  PrintText
END HandleEnter;

PROCEDURE HandleTextInput(key: INTEGER);
VAR ch: CHAR;
BEGIN
  IF G.IsPrintable(key) THEN
    ch := G.GetChar(key);
    IF ch # 0X THEN text.InsertChar(ch) END;
    text.selected := FALSE;
    PrintText
  ELSIF G.GetScancode(key) = G.kTab THEN
    text.InsertChar(' ');
    IF text.x MOD 2 # 0 THEN text.InsertChar(' ') END;
    text.selected := FALSE;
    PrintText
  END
END HandleTextInput;

PROCEDURE CountLines(s: ARRAY OF CHAR; width: INTEGER): INTEGER;
VAR i, x, lines: INTEGER;
BEGIN
  i := 0; x:= 0; lines := 1;
  WHILE s[i] # 0X DO
    IF s[i] = 0AX THEN
      INC(lines); x := 0
    ELSIF s[i] # 0DX THEN
      IF x = width - 1 THEN INC(lines); x := 0
      ELSE INC(x) END;
    END;
    INC(i)
  END;
  RETURN lines
END CountLines;

PROCEDURE PrintErrors(s: ARRAY OF CHAR);
VAR lines, width, x0, x, y, i: INTEGER;
BEGIN
  width := T.charsX - 2;
  lines := CountLines(s, width);
  IF lines > 10 THEN lines := 10 END;
  i := 0; x0 := 1; x := x0;
  y := T.charsY - 2 - lines;
  WHILE (s[i] # 0X) & (y < T.charsY - 2) DO
    IF s[i] = 0AX THEN
      WHILE x < x0 + width (* Till end of line *)
      DO T.PutChar(x, y, ' ', 0, 3); INC(x) END;
      x := x0; INC(y)
    ELSIF s[i] # 0DX THEN
      T.PutChar(x, y, s[i], 0, 3);
      IF x = x0 + width - 1 THEN INC(y); x := x0
      ELSE INC(x) END
    END;
    INC(i)
  END;
  IF x > x0 THEN
    WHILE x < x0 + width
    DO T.PutChar(x, y, ' ', 0, 3); INC(x) END
  END
END PrintErrors;

PROCEDURE QuickSave;
BEGIN
  IF text.SaveToFile(fileName) THEN
    fileIsNew := FALSE
  ELSE
    fileName := 'NOTSAVED' (*!FIXME*)
  END
END QuickSave;

PROCEDURE LoadFile;
BEGIN
  IF text.LoadFromFile(fileName) THEN
    fileIsNew := FALSE
  ELSE
    text.Clear;
    fileName := 'NONAME.Mod';
    fileIsNew := TRUE
  END
END LoadFile;

PROCEDURE DrawInputBox;
VAR i, x, y, w, h: INTEGER;
BEGIN
  w := 40; h := 10;
  x := (T.charsX - w) DIV 2;
  y := (T.charsY - h) DIV 2;
  FOR i := x + 1 TO x + w - 3 DO
    T.PutChar(i, y, 0CDX, 15, 7); (* = *)
    T.PutChar(i, y + h - 2, 0CDX, 15, 7);
    T.SetCharColor(i + 1, y + h - 1, 8, 0) (* Bottom shadow *)
  END;
  FOR i := y + 1 TO y + h - 2 DO
    T.PutChar(x, i, 0BAX, 15, 7); (* || *)
    T.PutChar(x + w - 3, i, 0BAX, 15, 7);
    T.SetCharColor(x + w - 2, i, 8, 0); (* Right shadow *)
    T.SetCharColor(x + w - 1, i, 8, 0)
  END;
  T.SetCharColor(x + w - 1, y + h - 1, 8, 0); (* Bottom-right shadow *)
  T.PutChar(x, y, 0C9X, 15, 7); (* Corners *)
  T.PutChar(x + w - 3, y, 0BBX, 15, 7);
  T.PutChar(x, y + h - 2, 0C8X, 15, 7);
  T.PutChar(x + w - 3, y + h - 2, 0BCX, 15, 7);
  T.CharFill(x + 1, y + 1, x + w - 4, y + h - 3, ' ', 0, 7);

  IF saveAs THEN
    T.PutString(x + w DIV 2 - 8, y, ' Save File As ', 15, 7, 0)
  ELSE
    T.PutString(x + w DIV 2 - 8, y, ' Open File As ', 15, 7, 0)
  END;
  T.PutString(x + 2, y, '[ ]', 15, 7, 0);
  T.PutChar(x + 3, y, 0FEX, 10, 7);

  IF saveAs THEN
    T.PutString(x + 3, y + 2, 'File to save', 15, 7, 0)
  ELSE
    T.PutString(x + 3, y + 2, 'File to open', 15, 7, 0)
  END;

  inputBox.x := x + 3; inputBox.y := y + 3;
  inputBox.w := w - 8; inputBox.Draw
END DrawInputBox;

PROCEDURE SaveAs;
BEGIN
  focusedWidget := inputBox;
  saveAs := TRUE;
  DrawInputBox
END SaveAs;

PROCEDURE HandleF1; (* Help *)
VAR w: INTEGER; box: Widgets.InfoBox;
BEGIN
  w := 42;
  box := Widgets.NewInfoBox((T.charsX - w) DIV 2 + 1, 3, w, T.charsY - 5);
  box.Draw;
  T.PutString(box.x + box.w DIV 2 - 4, box.y, ' Help ', 15, 3, 0);
  T.PutString(box.x + 2, box.y +  1, 'F2        - Save File', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  2, 'Shift+F2  - Save File As...', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  3, 'F3        - Open File...', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  4, 'Shift+F8  - Close File', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  5, 'F9        - Compile & Run', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  6, 'Shift+F9  - Compile, switch to', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  7, '            window mode & run', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  8, 'Alt+Enter - Toggle Fullscreen', 0, 3, 0);
  T.PutString(box.x + 2, box.y +  9, 'Shift+Esc - Exit Editor', 14, 3, 0);
  T.PutString(box.x + 2, box.y + 10, 'When saving,  write a file name like', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 11, '"MyProg.Mod".  Use extension  ".Mod"', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 12, 'with big "M".  Files are being saved', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 13, 'in subdirectory "Programs"  and com-', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 14, 'piled  with  Vishap Oberon Compiler.', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 15, 'FreeOberon version', 0, 3, 0);
  T.PutString(box.x + 21, box.y + 15, version, 1, 3, 0);
  T.PutString(box.x + 2, box.y + 16, '              Riga, December 4, 2016', 0, 3, 0);
  T.PutString(box.x + 2, box.y + 17, 'Press Escape to return to editor...', 0, 3, 0)
END HandleF1;

PROCEDURE HandleF2; (* Save *)
BEGIN
  IF fileIsNew OR keys[G.kLShift] OR keys[G.kRShift]
  THEN SaveAs
  ELSE QuickSave END
END HandleF2;

PROCEDURE HandleF3; (* Open *)
BEGIN
  focusedWidget := inputBox;
  saveAs := FALSE;
  DrawInputBox
END HandleF3;

PROCEDURE SwitchToEditor;
BEGIN
  state := stateEditor;
  PrintText
END SwitchToEditor;

PROCEDURE SwitchToTerminal;
BEGIN
  state := stateTerminal;
  inputBufLen := 0;
  T.CharFill(0, 0, T.charsX - 1, T.charsY - 1, ' ', 7, 0);
  T.GoToXY(0, 0)
END SwitchToTerminal;

PROCEDURE ParseErrors(VAR s: ARRAY OF CHAR);
VAR i, j, pos, st, len, skip: INTEGER; found: BOOLEAN;
BEGIN
  StringsFindNext(' pos ', s, 0, found, i);
  IF found THEN (* Read position *)
    WHILE (s[i] # 0X) &
          ((s[i] < '0') OR (s[i] > '9'))
    DO INC(i) END;
    IF (s[i] >= '0') & (s[i] <= '9') THEN
      pos := 0;
      REPEAT pos := pos * 10 + ORD(s[i]) - ORD('0'); INC(i)
      UNTIL (s[i] < '0') OR (s[i] > '9');
      (* Skip spaces before 'err' *)
      WHILE s[i] = ' ' DO INC(i) END;
      IF s[i] = 'e' THEN (* Assume 'err' reached *)
        skip := 3; (* Skip 3 characters *)
        WHILE (skip > 0) & (s[i] # 0X) DO INC(i); DEC(skip) END;
        WHILE s[i] = ' ' DO INC(i) END; (* Skip spaces *)
        WHILE (s[i] >= '0') & (s[i] <= '9') DO INC(i) END; (* Skip numbers *)
        WHILE s[i] = ' ' DO INC(i) END; (* Skip spaces *)
        s[0] := ' '; j := 1;
        WHILE s[i] >= ' ' DO s[j] := s[i]; INC(i); INC(j) END;
        (* Remove trailing spaces *)
        WHILE (j > 0) & (s[j] = ' ') DO DEC(i) END;
        s[j] := '.'; INC(j); s[j] := 0X;
        (* Capitalize first letter (0th is a space). *)
        IF (s[1] >= 'a') & (s[1] <= 'z') THEN s[1] := CAP(s[1]) END;
        text.MoveToPos(pos);
        PrintText
      END
    END
  END
END ParseErrors;

PROCEDURE Compile(): BOOLEAN;
CONST bufLen = 20480;
VAR buf: ARRAY bufLen OF CHAR;
    len, err: INTEGER;
    cmd: ARRAY 1024 OF CHAR;
    s, sN: ARRAY 80 OF CHAR;
    success: BOOLEAN;
BEGIN
  IF Config.isWindows THEN
    IF Term.SearchPath('cmd.exe', cmd) # 0 THEN
      Strings.Insert('"', 0, cmd);
      Strings.Append('" /C data\bin\compile.bat ', cmd);
    ELSE T.PutString(0, T.charsY - 1, 'Could not find cmd.exe', 15, 4, 0)
    END
  ELSE (* Linux *)
    COPY('data/bin/compile.sh ', cmd)
  END;
  Strings.Append(fileName, cmd);
  success := (Term.RunProcess(cmd, buf, bufLen, len, err) # 0) &
             (err = 0);
  IF ~success THEN
    COPY(' Compilation returned ', s);
    IntToStr(err, sN);
    Strings.Append(sN, s);
    Strings.Append(' exit status ', s);
    T.PutString(0, T.charsY - 1, s, 15, 4, 0);
    IF (len > 0) & (len < bufLen) THEN
      IF buf[len - 1] = 0AX THEN buf[len - 1] := 0X
      ELSE buf[len] := 0X
      END;
      ParseErrors(buf)
    ELSE COPY(' Compilation failed.', buf)
    END;
    PrintErrors(buf)
  END;
  RETURN success
END Compile;

PROCEDURE RunProgram;
VAR cmd, prg: ARRAY 128 OF CHAR;
    x: INTEGER;
BEGIN
  (* Extract 'MyProg' from 'MyProg.Mod' *)
  COPY(fileName, prg);
  x := Strings.Length(prg);
  WHILE (x > 0) & (prg[x] # '.') DO DEC(x) END;
  IF prg[x] = '.' THEN prg[x] := 0X END;

  (* Construct 'bin/MyProg' or 'bin\MyProg' *)
  IF Config.isWindows THEN COPY('bin\', cmd) ELSE COPY('bin/', cmd) END;
  Strings.Append(prg, cmd);

  IF ~Term.StartProcess(cmd) THEN
    T.PutString(0, T.charsY - 1, ' Program execution failed ', 15, 4, 0)
  ELSE
    programFinished := FALSE;
    SwitchToTerminal
  END
END RunProgram;

PROCEDURE HandleShiftF8;
BEGIN
  text.Clear;
  fileIsNew := TRUE;
  fileName := 'NONAME.Mod';
  PrintText
END HandleShiftF8;

(* Compile & Run.
   If needWindowed is true and editor is in fullscreen
   mode, then editor will temporary switch to window
   mode before running the program. *)
PROCEDURE HandleF9(needWindowed: BOOLEAN);
BEGIN
  IF fileIsNew THEN
    SaveAs
  ELSE
    QuickSave;
    IF Compile() THEN
      tempWindowed := needWindowed & isFullscreen;
      IF tempWindowed THEN G.SwitchToWindowed END;
      RunProgram
    END
  END
END HandleF9;

PROCEDURE HandleAltEnter; (* Toggle Fullscreen *)
BEGIN
  IF isFullscreen THEN G.SwitchToWindowed
  ELSE G.SwitchToFullscreen END;
  isFullscreen := ~isFullscreen
END HandleAltEnter;

PROCEDURE HandleEditorWindowKeyDown(key: INTEGER; VAR quit: BOOLEAN);
VAR code: INTEGER;
BEGIN
  code := G.GetScancode(key);
  CASE code OF
    G.kEsc:
    IF keys[G.kLShift] OR keys[G.kRShift] THEN quit := TRUE
    ELSE PrintText
    END
  | G.kLeft:      MoveInLine(FALSE)
  | G.kRight:     MoveInLine(TRUE)
  | G.kUp:        MoveByLine(FALSE, FALSE)
  | G.kDown:      MoveByLine(TRUE, FALSE)
  | G.kHome:      MoveToLineEdge(FALSE)
  | G.kEnd:       MoveToLineEdge(TRUE)
  | G.kPgUp:      MoveByPage(FALSE)
  | G.kPgDn:      MoveByPage(TRUE)
  | G.kBackspace: HandleBackspace
  | G.kDel:       HandleDelete
  | G.kEnter, G.kEnterPad:
    IF keys[G.kAlt] OR keys[G.kAltGr]
    THEN HandleAltEnter ELSE HandleEnter END
  | G.kF1: HandleF1
  | G.kF2: HandleF2
  | G.kF3: HandleF3
  | G.kF8:
    IF keys[G.kLShift] OR keys[G.kRShift]
    THEN HandleShiftF8 END
  | G.kF9:
    IF keys[G.kLCtrl] OR keys[G.kRCtrl]
    THEN HandleF9(TRUE) ELSE HandleF9(FALSE) END
  ELSE
    IF G.ControlPressed() THEN
      CASE code OF 
        G.kX: CutText
      | G.kC: CopyText
      | G.kV: PasteText
      | G.kDel, G.kDelPad: DeleteText
      ELSE HandleTextInput(key) END
    ELSE HandleTextInput(key) END
  END
END HandleEditorWindowKeyDown;

PROCEDURE CloseInputBox;
BEGIN
  focusedWidget := editor;
  PrintText
END CloseInputBox;

PROCEDURE HandleEditorKeyDown(key: INTEGER; VAR quit: BOOLEAN);
VAR code: INTEGER;
BEGIN
  IF focusedWidget = editor THEN
    HandleEditorWindowKeyDown(key, quit)
  ELSIF focusedWidget = inputBox THEN
    CASE G.GetScancode(key) OF
      G.kEsc: CloseInputBox
    | G.kEnter, G.kEnterPad:
      CloseInputBox;
      COPY(inputBox.s, fileName);
      IF saveAs THEN QuickSave ELSE LoadFile END;
      PrintText
    ELSE
      focusedWidget.HandleKeyDown(key);
      focusedWidget.Draw
    END
  END
END HandleEditorKeyDown;

PROCEDURE HandleMouseMotion;
VAR x, y, newX, newY: INTEGER;
BEGIN
  G.GetMousePos(x, y);
  newX := x DIV T.charW;  newY := y DIV T.charH;
  IF (newX # T.mouseX) OR (newY # T.mouseY) THEN T.MouseXY(newX, newY) END
END HandleMouseMotion;

PROCEDURE PollProgram;
VAR len, i: INTEGER;
    err: INTEGER;
    s, sN: ARRAY 64 OF CHAR;
  PROCEDURE WriteProgBuf;
  VAR ch: CHAR; i: INTEGER;
  BEGIN
    i := 0;
    WHILE i < len DO
      ch := progBuf[i];
      IF ch = 0D0X THEN
        INC(i); ch := progBuf[i];
        IF ch = 081X THEN ch := CHR(240) (* Big Yo *)
        ELSE ch := CHR(ORD(ch) - 16)
        END
      ELSIF ch = 0D1X THEN
        INC(i); ch := CHR(ORD(progBuf[i]) + 96)
      ELSIF ch >= 080X THEN ch := '?'
      END;
      T.Write(ch); INC(i)
    END
  END WriteProgBuf;

  PROCEDURE Read(tillEnd: BOOLEAN);
  VAR loopLimit: INTEGER;
  BEGIN
    loopLimit := 5;
    REPEAT
      Term.ReadFromProcess(progBuf, len, LEN(progBuf));
      IF len > 0 THEN
        IF inputBufLen > 0 THEN
          FOR i := 0 TO inputBufLen - 1 DO T.Backspace END;
          inputBufLen := 0
        END;
        WriteProgBuf
      END;
      DEC(loopLimit)
    UNTIL (len <= 0) OR (loopLimit <= 0) & ~tillEnd
  END Read;
BEGIN
  IF ~programFinished THEN
    IF Term.ProcessFinished(err) THEN
      Read(TRUE); (* Read everything until pipe is empty *)
      programFinished := TRUE;
      IF tempWindowed THEN G.SwitchToFullscreen END;
      IF err = 0 THEN
        T.WriteString(' Press any key to return to IDE')
      ELSE
        s := ' Runtime error ';
        IntToStr(err, sN); Strings.Append(sN, s);
        T.WriteString(s)
      END
    ELSE
      Read(FALSE) (* Attempt several reads *)
    END
  END
END PollProgram;

PROCEDURE WriteToProcess(s: ARRAY OF CHAR; len: INTEGER);
VAR buf: ARRAY 2048 OF CHAR; i, bufLen: INTEGER; ch: CHAR;
BEGIN
  bufLen := 0; i := 0;
  WHILE i < len DO
    ch := s[i];
    IF ch < 80X THEN
      buf[bufLen] := ch; INC(bufLen)
    ELSIF ORD(ch) = 240 THEN (* Big cyrillic Yo *)
      buf[bufLen] := 0D0X; buf[bufLen + 1] := 81X;
      INC(bufLen, 2)
    ELSIF ORD(ch) < 224 THEN (* Before small cyrillic R *)
      buf[bufLen] := 0D0X;
      buf[bufLen + 1] := CHR(ORD(ch) - 128 + 090H);
      INC(bufLen, 2)
    ELSE
      buf[bufLen] := 0D1X;
      buf[bufLen + 1] := CHR(ORD(ch) - 224 + 080H);
      INC(bufLen, 2)
    END;
    INC(i)
  END;
  Term.WriteToProcess(buf, bufLen)
END WriteToProcess;

PROCEDURE HandleTerminalKeyDown(key: INTEGER);
VAR code: INTEGER; ch: CHAR; buf: ARRAY 2 OF CHAR;
BEGIN
  IF programFinished THEN SwitchToEditor
  ELSE
    CASE G.GetScancode(key) OF
      G.kEnter, G.kEnterPad:
      T.Ln;
      WriteToProcess(inputBuf, inputBufLen);
      inputBufLen := 0; buf[0] := 0AX;
      Term.WriteToProcess(buf, 1)
    | G.kBackspace:
      IF (inputBufLen > 0) THEN
        DEC(inputBufLen);
        T.Backspace
      END
    | G.kPause:
      IF G.ControlPressed() THEN
        programFinished := TRUE;
        SwitchToEditor (* !FIXME Kill the process *)
      END
    ELSE
      IF G.IsPrintable(key) THEN
        ch := G.GetChar(key);
        IF (ch # 0X) & (inputBufLen < LEN(inputBuf)) THEN
          inputBuf[inputBufLen] := ch; INC(inputBufLen);
          T.Write(ch)
        END
      END
    END
  END
END HandleTerminalKeyDown;

PROCEDURE Run;
VAR quit: BOOLEAN;
    event: G.Event;
BEGIN
  quit := FALSE;
  REPEAT
    CASE state OF
      stateEditor:
      WHILE G.PollEvent(event) DO
        CASE event.type OF
          G.mouseMove: HandleMouseMotion
        | G.keyDown: HandleEditorKeyDown(event.key, quit)
        | G.quit: quit := TRUE
        ELSE
        END
      END
    | stateTerminal:
      WHILE G.PollEvent(event) DO
        CASE event.type OF
          G.mouseMove: HandleMouseMotion
        | G.keyDown: HandleTerminalKeyDown(event.key)
        ELSE
        END
      END;
      PollProgram
    ELSE
    END;
    T.Act;
    T.Draw;
    G.Flip
  UNTIL quit
END Run;

PROCEDURE InitIDE;
BEGIN
  state := stateEditor;
  text := Text.NewText();
  fileIsNew := TRUE; fileName[0] := 0X;
  fileName := 'NONAME.Mod';
  clipboard[0] := 0X;

  editor := Widgets.NewEditingWindow(0, 1, T.charsX, T.charsY - 2);
  inputBox := Widgets.NewEdit(10, 10, 20, 1);
  focusedWidget := editor;

  PrintIDE
END InitIDE;

PROCEDURE Init(): BOOLEAN;
VAR success: BOOLEAN;
BEGIN
  success := FALSE;
  isFullscreen := needFullscreen;
  IF T.Init(isFullscreen) THEN
    keys := G.GetKeyArray();
    InitIDE;
    success := TRUE
  ELSE Out.String('Terminal init failed.'); Out.Ln
  END;
  RETURN success
END Init;

BEGIN
  IF ~Init() THEN
    Out.String('Could not initialize.'); Out.Ln
  ELSE
    PrintText;
    Run
  END;
  G.Close
END FreeOberon.
